/*
 * Copyright (C) 2023 Cristian Sabio
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.cristian.libtxz;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * This class implements some utilities to use with LibTXZ library.
 * @author Cristian
 */
public class TXZUtils {
    private TXZUtils() {}
    
    /**
     * Returns true if in the given path exists an element, false otherwise.
     * @param ElementPath String containing the path of the element.
     * @return Boolean true value if the element exists, false otherwise.
     * @see isFile
     * @see isDirectory
     */
    public static boolean exists(String ElementPath) {
        File FileExistsObject = new File(ElementPath);
        return FileExistsObject.exists();
    }
    
    /**
     * Like the name implies, this method returns true if the element is a file.
     * @param ElementPath String containing the path of the element.
     * @return Boolean true if is a file, false otherwise.
     * @see isDirectory
     */
    public static boolean isFile(String ElementPath) {
        File FileType = new File(ElementPath);
        return FileType.isFile();
    }
    
    /**
     * Like the name implies, this method returns true if the element is a directory.
     * @param ElementPath String containing the path of the element.
     * @return Boolean true if is a directory, false otherwise.
     * @see isFile
     */
    public static boolean isDirectory(String ElementPath) {
        File DirectoryType = new File(ElementPath);
        return DirectoryType.isDirectory();
    }
    
    /**
     * Scans the given directory and returns a File array with all the files and
     * directories found in.
     * @param DirectoryToScan String containing the path of the directory.
     * @param RecursiveScan Boolean value indicating a recursive scan.
     * @param Levels Integer containing the level of recursion, if this is -1,
     * it recurses infinitely.
     * @return File array with all of the files and directories inside the directory.
     * @throws FileNotFoundException Throwed when the directory doesn't exists.
     * @throws IOException Throwed when the given path is not a directory.
     */
    public static File[] scanDirectory(String DirectoryToScan, boolean RecursiveScan, int Levels) throws FileNotFoundException, IOException {
        if(exists(DirectoryToScan)) {
            if(isDirectory(DirectoryToScan)) {
                ArrayList<File> DirectoryTree = new ArrayList<>();
                File Directory = new File(DirectoryToScan);

                for(String Element : Directory.list()) {
                    File ElementObject = new File(Directory.getAbsolutePath() + File.separator + Element);
                    DirectoryTree.add(ElementObject);
                    if(ElementObject.isDirectory() && RecursiveScan && (Levels > 0 || Levels == -1)) {
                        File[] SubListing = TXZUtils.scanDirectory(Directory.getAbsolutePath() + File.separator + Element, true, Levels == -1 ? Levels : Levels - 1);
                        DirectoryTree.addAll(Arrays.asList(SubListing));
                    }
                }
        
                File[] FinalTree = new File[DirectoryTree.size()];
                for(int I = 0; I < DirectoryTree.size(); I++) {
                    FinalTree[I] = DirectoryTree.get(I);
                }
                return FinalTree;
            } else {
                throw new IOException("There is not a directory in the supplied path!");
            }
        } else {
            throw new FileNotFoundException("The given path doesn't exists!");
        }
    }
    
    /**
     * Scans the given directory and returns a File array with all the files and
     * directories found in. This method doesn't scan recursively.
     * @param DirectoryToScan String containing the path of the directory.
     * @return File array with all of the files and directories inside the directory.
     * @throws FileNotFoundException Throwed when the directory doesn't exists.
     * @throws IOException Throwed when the given path is not a directory.
     */
    public static File[] scanDirectory(String DirectoryToScan) throws FileNotFoundException, IOException {
        return scanDirectory(DirectoryToScan, false, -1);
    }
}
