/*
 * Copyright (C) 2023 Cristian Sabio
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.cristian.txz;

import org.cristian.libtxz.*;
import org.kamranzafar.jtar.TarEntry;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.HelpFormatter;
import java.io.File;
import java.nio.file.Paths;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ByteArrayOutputStream;
import org.apache.commons.cli.ParseException;
import java.io.IOException;

/**
 * This is the TXZ CLI program.
 * @author Cristian
 */
public class ProgramCLI {
    private static final int DEFAULT_COMPRESSION_LEVEL = 7;
    
    private ProgramCLI() {}
    
    public static void main(String[] args) {
        try {
            CommandLine argv = parseOptions(args);
            if(argv.hasOption("help")) printHelp();
            else if(argv.hasOption("license")) printLicenses();
            else if(argv.hasOption("create")) create(argv.getOptionValues("create"), argv.hasOption("file") ? new File(argv.getOptionValue("file").endsWith(".txz") ? argv.getOptionValue("file") : argv.getOptionValue("file") + ".txz") : null, argv.hasOption("level") ? Integer.parseInt(argv.getOptionValue("level")) : DEFAULT_COMPRESSION_LEVEL, argv.hasOption("verbose"));
            else if(argv.hasOption("list")) list(argv.hasOption("file") ? new File(argv.getOptionValue("file")) : null, argv.hasOption("verbose"));
            else if(argv.hasOption("extract")) extract(argv.hasOption("file") ? new File(argv.getOptionValue("file")) : null, argv.hasOption("output") ? new File(argv.getOptionValue("output")) : null, argv.hasOption("verbose"));
            else if(argv.hasOption("checksums")) computeChecksums(argv.hasOption("file") ? new File(argv.getOptionValue("file")) : null, argv.hasOption("output") ? new File(argv.getOptionValue("output")) : null, argv.hasOption("verbose"));
            else System.err.println("No such option. Use the -h option for help."); System.exit(1);
            System.exit(0);
        } catch(Throwable Error) {
            System.err.println("Unexpected error: " + Error.toString());
            System.exit(-1);
        }
    }
    
    private static Options createOptions() {
        Option CreateOpt = new Option("c", "create", true, "Creates a file, if no -f option is present, the program will write to the standard output.");
        CreateOpt.setArgs(Option.UNLIMITED_VALUES);
        Option[] Opts = {
            new Option("h", "help", false, "Prints the help text."),
            new Option("w", "license", false, "Prints the full program license. I recomend that you use the command 'more' because this option will print all of the licenses of this program."),
            CreateOpt,
            new Option("t", "list", false, "List a file's contents, if no -f option is present, this program will read from standard input."),
            new Option("x", "extract", false, "Extract a file, if no -f is present, the program will read from standard input."),
            new Option("s", "checksums", false, "Compute all the checksums of the files contained in a file, if no -f option is present, this program will read from standard input, also if no -o option is present, the program will write to standard output."),
            new Option("l", "level", true, "Set the compression level, default is 7."),
            new Option("f", "file", true, "Set the output file."),
            new Option("o", "output", true, "Set the output directory, if the directory doesn't exists, txz will create one."),
            new Option("v", "verbose", false, "Prints more information during run."),
        };
        
        Options ProgramOptions = new Options();
        
        for(Option Element : Opts) {
            ProgramOptions.addOption(Element);
        }
        
        return ProgramOptions;
    }
    
    private static CommandLine parseOptions(String[] argv) {
        Options           Opts       = createOptions();
        CommandLineParser OptsParser = new DefaultParser();
        @SuppressWarnings("unused")
        CommandLine       OptsCmd    = null;
        try {
            OptsCmd = OptsParser.parse(Opts, argv);
        } catch(ParseException Error) {
            System.err.println("Unable to parse the command line options: " + Error.toString());
            System.exit(2);
        }
        return OptsCmd;
    }
    
    private static void printHelp() {
        new HelpFormatter().printHelp(
                "txz",
                "Handle Tar+XZ files.",
                createOptions(),
                "\nThis program is free software licensed under the General Public License Version 3, you can read more using the -w option.",
                true
        );
    }
    
    protected static void printResource(String Path) {
        try(InputStream In = ProgramCLI.class.getResourceAsStream(Path)) {
            int Data;
            ByteArrayOutputStream Buffer = new ByteArrayOutputStream();
            while((Data = In.read()) != -1) Buffer.write(Data);
            System.out.println(new String(Buffer.toByteArray()));
        } catch(NullPointerException | IOException Error) {
            System.err.println("Unable to read resource '" + Path + "': " + Error.toString());
            System.exit(3);
        }
    }
    
    protected static void printLicenses() {
        System.out.println("LibTXZ License:");
        printResource("/org/cristian/libtxz/LICENSE");
        System.out.println("\nApache Commons CLI License:");
        printResource("/org/apache/commons/cli/LICENSE");
        System.out.println("\nJTar License:");
        printResource("/org/kamranzafar/jtar/LICENSE.txt");
        System.out.println("\nXZ License:");
        printResource("/org/tukaani/xz/COPYING");
    }
    
    protected static void create(String[] InputList, File Output, int Level, boolean Verbose) {
        Level = Level >= 0 && Level <= 9 ? Level : DEFAULT_COMPRESSION_LEVEL;
        try {
            OutputStream Out = Output == null ? System.out : new FileOutputStream(Output);
            try(TXZOutputStream OutTXZ = new TXZOutputStream(Out, Level)) {
                for(String InputElement : InputList) {
                    File InputFile = new File(InputElement);
                    OutTXZ.putNextEntry(new TXZEntry(InputFile, InputFile.getAbsolutePath().replace(InputFile.getAbsoluteFile().getParent() + File.separator, "")));
                    if(InputFile.isDirectory()) {
                        if(Verbose && Output == null)      System.err.println(InputFile.getAbsolutePath());
                        else if(Verbose && Output != null) System.out.println(InputFile.getAbsolutePath());
                        for(File Input : TXZUtils.scanDirectory(InputFile.getAbsolutePath(), true, -1)) {
                            OutTXZ.putNextEntry(new TXZEntry(Input, Input.getAbsolutePath().replace(InputFile.getAbsoluteFile().getParent() + File.separator, "")));
                            if(Input.isFile()) {
                                try(InputStream InputFileStream = new FileInputStream(Input)) {
                                    int Data;
                                    while((Data = InputFileStream.read()) != -1) OutTXZ.write(Data);
                                }
                            }
                            if(Verbose && Output == null)      System.err.println(Input.getAbsolutePath());
                            else if(Verbose && Output != null) System.out.println(Input.getAbsolutePath());
                        }
                    } else {
                        try(InputStream InputFileStream = new FileInputStream(InputFile)) {
                            int Data;
                            while((Data = InputFileStream.read()) != -1) OutTXZ.write(Data);
                        }
                        if(Verbose && Output == null)      System.err.println(InputFile.getAbsolutePath());
                        else if(Verbose && Output != null) System.out.println(InputFile.getAbsolutePath());
                    }
                }
            }
            if(Output != null) Out.close();
        } catch(IOException Error) {
            System.err.println("File creation error: " + Error.toString());
            if(Verbose) Error.printStackTrace(System.err);
            System.exit(10);
        }
    }
    
    protected static void list(File Input, boolean Verbose) {
        try {
            InputStream In = Input == null ? System.in : new FileInputStream(Input);
            try(TXZInputStream InTXZ = new TXZInputStream(In)) {
                TarEntry CurrentEntry;
                while((CurrentEntry = InTXZ.getNextEntry()) != null) System.out.println((Verbose ? "[" + CurrentEntry.getModTime() + "]\t[" + (CurrentEntry.isDirectory() ? "D": "F") + "] [" + String.format("%10d", CurrentEntry.getSize()).replace(" ", "-") + "] " : "") + CurrentEntry.getName());
            }
            if(Input != null) In.close();
        } catch(IOException Error) {
            System.err.println("File reading error: " + Error.toString());
            if(Verbose) Error.printStackTrace(System.err);
            System.exit(20);
        }
    }
    
    protected static void extract(File Input, File Output, boolean Verbose) {
        try {
            InputStream In = Input == null ? System.in : new FileInputStream(Input);
            Output = Output == null ? Paths.get(ProgramCLI.class.getProtectionDomain().getCodeSource().getLocation().getPath()).getParent().toFile() : Output;
            if(!Output.exists()) Output.mkdirs();
            try(TXZInputStream InTXZ = new TXZInputStream(In)) {
                TarEntry CurrentEntry;
                while((CurrentEntry = InTXZ.getNextEntry()) != null) {
                    File CurrentFile = new File(Output.getAbsolutePath() + File.separator + CurrentEntry.getName());
                    if(CurrentEntry.isDirectory()) {
                        CurrentFile.mkdirs();
                    } else {
                        try(OutputStream CurrentEntryOutput = new FileOutputStream(Output.getAbsolutePath() + File.separator + CurrentEntry.getName())) {
                            int Data;
                            while((Data = InTXZ.read()) != -1) CurrentEntryOutput.write(Data);
                        }
                    }
                    CurrentFile.setLastModified(CurrentEntry.getModTime().getTime());
                    if(Verbose) System.out.println(Output.getAbsolutePath() + File.separator + CurrentEntry.getName());
                }
            }
            if(Input != null) In.close();
        } catch(IOException Error) {
            System.err.println("File extraction error: " + Error.toString());
            if(Verbose) Error.printStackTrace(System.err);
            System.exit(30);
        }
    }
    
    protected static void computeChecksums(File Input, File Output, boolean Verbose) {
        try {
            InputStream  In  = Input  == null ? System.in : new FileInputStream(Input);
            OutputStream Out = Output == null ? System.out : new FileOutputStream(Output);
            try(TXZInputStream InTXZ = new TXZInputStream(In)) {
                TarEntry CurrentEntry;
                while((CurrentEntry = InTXZ.getNextEntry()) != null) {
                    try(ByteArrayOutputStream DataStream = new ByteArrayOutputStream()) {
                        int Data;
                        while((Data = InTXZ.read()) != -1) DataStream.write(Data);
                        Out.write((String.format("%016X", CurrentEntry.computeCheckSum(DataStream.toByteArray())) + "  " + CurrentEntry.getName() + "\n").getBytes());
                    }
                }
            }
            if(Input != null)  In.close();
            if(Output != null) Out.close();
        } catch(IOException Error) {
            System.err.println("File reading error: " + Error.toString());
            if(Verbose) Error.printStackTrace(System.err);
            System.exit(40);
        }
    }
}
